import logging

import shovl.schemas as schemas

logger = logging.getLogger("shovl.providers")


class ServiceProvider:
    """
    Base class for service providers. Inherited by all service providers
    (e.g. database, bucket, etc.) to provide common functionality.
    """

    config: schemas.ServiceConfig

    def __init__(self, config: schemas.ServiceConfig) -> None:
        """
        Initialize the service provider.
        """
        self.config = config
        self._cleanup()

    def _cleanup(self) -> None:
        """
        Cleanup the service provider's state.
        """
        raise NotImplementedError

    @property
    def connected(self) -> bool:
        """
        Check if the provider is connected to the service.
        """
        raise NotImplementedError

    def connect(self) -> None:
        """
        Connect to the service.
        """
        raise NotImplementedError

    def disconnect(self) -> None:
        """
        Disconnect from the service.
        """
        raise NotImplementedError



import datetime
import enum
from typing import Optional

import pydantic

from .service import ServiceEntity


class BucketEntityType(enum.Enum):
    """
    The type of an entity in the bucket.
    """

    bucket = "Bucket"
    file = "File"
    folder = "Folder"


class BucketEntity(ServiceEntity):
    """
    Represents an entity in a bucket.
    """

    last_modified: Optional[datetime.datetime] = pydantic.Field(
        default=None,
        description="Last modified timestamp of the entity (only for files).",
    )
    name: str = pydantic.Field(description="Name of the entity.")
    size: Optional[int] = pydantic.Field(
        default=None,
        description="Size of the entity in bytes (only for files).",
    )
    type: BucketEntityType = pydantic.Field(description="Type of the entity.")



import datetime

import pydantic


class TemporaryCredentials(pydantic.BaseModel):
    """
    Temporary AWS credentials, typically obtained from AWS STS (Security Token
    Service).
    """

    access_key_id: str = pydantic.Field(
        ...,
        alias="AccessKeyId",
        description="The access key ID of the temporary credentials.",
    )
    expiration: datetime.datetime = pydantic.Field(
        ...,
        alias="Expiration",
        description="The expiration time of the temporary credentials.",
    )
    secret_access_key: str = pydantic.Field(
        ...,
        alias="SecretAccessKey",
        description="The secret access key of the temporary credentials.",
    )
    session_token: str = pydantic.Field(
        ...,
        alias="SessionToken",
        description="The session token of the temporary credentials.",
    )



from .service import ServiceProvider

__all__ = ["ServiceProvider"]

# Database and Bucket providers are imported lazily in the connection screen due
# to optional dependencies on database and S3 providers. This allows the app
# to run without installing database and S3 dependencies if those features are
# not needed.



import logging
from typing import List, Sequence

from textual.app import ComposeResult
from textual.widgets import DataTable, Footer, Header

import shovl.providers as providers
import shovl.schemas as schemas

from .base import BaseScreen

logger = logging.getLogger("shovl.screens.service")


class ServiceScreen(BaseScreen):
    """
    Base service screen. Inherited by all service screens (e.g. database,
    bucket, etc.) to provide common functionality such as displaying
    connection info, handling common keybindings, etc.
    """

    CSS_PATH = "../styles/service.tcss"

    table: DataTable

    connection: schemas.ServiceConfig
    provider: providers.ServiceProvider

    history: List[schemas.ServiceEntity]
    all_entities: Sequence[schemas.ServiceEntity]
    filtered_entites: Sequence[schemas.ServiceEntity]

    current_entity: schemas.ServiceEntity | None

    ### Lifecycle methods ###

    def __init__(
        self,
        config: schemas.ConfigSchema,
        connection: schemas.ServiceConfig,
        provider: providers.ServiceProvider,
    ) -> None:
        """
        Initialize the database screen.
        """
        super().__init__(config=config)
        self.connection = connection
        self.provider = provider

        self.history = []
        self.all_entities = []
        self.filtered_entities = []

        self.current_entity = None

        self.set_title(f"shovl - {self.connection.name}")

    def compose(self) -> ComposeResult:
        """
        Compose the service screen.
        """
        yield Header()
        self.table = DataTable(
            classes="table",
            id="table",
        )
        yield self.table
        yield Footer()

    ### Actions ###

    def action_back(self) -> None:
        """
        Go back one step in history, if available.
        """
        previous_entity = self.go_back_in_history()
        if previous_entity:
            self.run_in_thread(
                target=self.inspection_worker,
                name="InspectionWorker",
                entity=previous_entity,
                add_current_to_history=False,
            )
        else:
            self.provider.disconnect()
            self.app.pop_screen()

    ### Class methods ###

    def focus_view(self) -> None:
        """
        Focus the tabe.
        """
        self.table.focus()

    def add_columns(self, entity: schemas.ServiceEntity) -> None:
        """
        Add columns to the table based on the type of entity.
        """
        raise NotImplementedError(
            "add_columns must be implemented by subclasses"
        )

    def add_row(self, entity: schemas.ServiceEntity) -> None:
        """
        Add a row to the table.
        """
        raise NotImplementedError("add_rows must be implemented by subclasses")

    def refresh_view(self, show_truncated_hint: bool) -> None:
        """
        Add rows to the table based on a list of entities.
        """
        with self.app.batch_update():
            self.table.clear(columns=True)

            if not self.filtered_entities:
                super().refresh_view(show_truncated_hint=show_truncated_hint)
                return

            self.add_columns(entity=self.filtered_entities[0])

            for index, entity in enumerate(self.filtered_entities):
                if (
                    index >= self.config.gui.max_list_items
                    and self.config.gui.max_list_items > 0
                ):
                    if show_truncated_hint:
                        self.notify_info(
                            f"Showing first {self.config.gui.max_list_items} "
                            f"items. Refine your filter to see more."
                        )
                    break

                self.add_row(entity=entity)

        super().refresh_view(show_truncated_hint=show_truncated_hint)

    def get_selected_row(self) -> schemas.ServiceEntity | None:
        """
        Get the data of the currently selected row in the table and return it
        as a BucketEntity object.
        """
        if self.table.cursor_row is None:
            return None

        if self.table.cursor_row >= len(self.filtered_entities):
            return None

        return self.filtered_entities[self.table.cursor_row]

    def add_to_history(self, entity: schemas.ServiceEntity) -> None:
        """
        Add an item to the history stack.
        """
        self.history.append(entity)

    def go_back_in_history(self) -> schemas.ServiceEntity | None:
        """
        Go back in history and return the previous entity, or None if at root.
        """
        if len(self.history) >= 1:
            return self.history.pop()
        elif len(self.history) == 0:
            return None

    def inspect(
        self, entity: schemas.ServiceEntity, add_current_to_history: bool
    ) -> None:
        """
        Inspect a bucket entity and update the view with the results. This
        method is meant to be called from a separate thread to avoid blocking
        the UI.
        """
        raise NotImplementedError("inspect must be implemented by subclasses")

    ### Workers ###

    def inspection_worker(
        self, entity: schemas.BucketEntity, add_current_to_history: bool
    ) -> None:
        """
        Inspect a service entity.
        """
        try:
            self.app.call_from_thread(
                self.push_status_screen, label="Inspecting..."
            )

            self.inspect(
                entity=entity, add_current_to_history=add_current_to_history
            )

            self.app.call_from_thread(
                self.refresh_view, show_truncated_hint=True
            )

            self.app.call_from_thread(self.dismiss_status_screen)
        except Exception as exception:
            logger.exception(f"Inspection failed: {exception}")
            self.app.call_from_thread(self.dismiss_status_screen)
            self.app.call_from_thread(self.focus_view)
            message = str(exception)
            self.app.call_from_thread(
                lambda: self.notify_error(
                    f"Unable to inspect entity: {message}"
                )
            )



from typing import Callable, Optional

from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.timer import Timer
from textual.widgets import Input


class FilterScreen(ModalScreen):
    """
    Screen to filter a list of items.
    """

    BINDINGS = [("escape", "cancel", "cancel")]

    CSS_PATH = "../styles/filter.tcss"

    change_callback: Callable[[str], None]
    confirm_callback: Callable[[str], None]
    cancel_callback: Callable[[str], None]

    current_filter: str

    debounce_timer: Optional[Timer]

    ### Lifecycle methods ###

    def __init__(
        self,
        current_filter: str = "",
        change_callback: Callable[[str], None] = lambda x: None,
        confirm_callback: Callable[[str], None] = lambda x: None,
        cancel_callback: Callable[[str], None] = lambda x: None,
    ) -> None:
        """
        Initialize the filter screen.
        """
        super().__init__()

        self.change_callback = change_callback
        self.confirm_callback = confirm_callback
        self.cancel_callback = cancel_callback
        self.current_filter = current_filter
        self.debounce_timer = None

    def compose(self) -> ComposeResult:
        """
        Compose the filter screen.
        """
        yield Input(
            classes="input",
            placeholder="Type to filter...",
            value=self.current_filter,
        )

    ### Actions ###

    def action_cancel(self) -> None:
        """
        Cancel the filter operation.
        """
        input = self.query_one(Input)

        if self.debounce_timer:
            self.debounce_timer.stop()
            self.debounce_timer = None

        self.cancel_callback(input.value)

    ### Event handlers ###

    def on_input_changed(self, event: Input.Changed) -> None:
        """
        Handle filter input changes with debouncing.
        """
        if self.debounce_timer:
            self.debounce_timer.stop()

        self.debounce_timer = self.set_timer(
            0.5,
            lambda: self.change_callback(event.value),
        )

    def on_input_submitted(self, event: Input.Submitted) -> None:
        """
        Handle filter input submission.
        """
        if self.debounce_timer:
            self.debounce_timer.stop()
            self.debounce_timer = None

        self.confirm_callback(event.value)



from .aws import TemporaryCredentials
from .bucket import BucketEntity, BucketEntityType
from .config import (
    BucketConfig,
    ConfigSchema,
    DatabaseConfig,
    GuiConfig,
    LoggingConfig,
    ProviderConfig,
    ServiceConfig,
)
from .database import DatabaseEntity, DatabaseEntityType
from .path import PathRestriction
from .service import ServiceEntity

__all__ = [
    "TemporaryCredentials",
    "BucketEntity",
    "BucketEntityType",
    "BucketConfig",
    "ConfigSchema",
    "ServiceConfig",
    "DatabaseConfig",
    "GuiConfig",
    "LoggingConfig",
    "ProviderConfig",
    "DatabaseEntity",
    "DatabaseEntityType",
    "PathRestriction",
    "ServiceEntity",
]



from .base import BaseScreen
from .connection import ConnectionScreen
from .filter import FilterScreen
from .path import PathScreen
from .service import ServiceScreen
from .status import StatusScreen

__all__ = [
    "BaseScreen",
    "ConnectionScreen",
    "FilterScreen",
    "PathScreen",
    "StatusScreen",
    "ServiceScreen",
]

# Database and S3 screens are imported lazily in the connection screen due
# to optional dependencies on database and S3 providers. This allows the app
# to run without installing database and S3 dependencies if those features are
# not needed.



import logging

from textual.app import App

import shovl.schemas as schemas
import shovl.screens as screens

logger = logging.getLogger("shovl.app")


class ShovlApp(App):
    """
    The main application class for Shovl.
    """

    ENABLE_COMMAND_PALETTE = False

    config: schemas.ConfigSchema

    def __init__(self, config: schemas.ConfigSchema) -> None:
        """
        Initialize the Shovl application with the given configuration.
        """
        super().__init__()
        self.config = config

    def on_mount(self) -> None:
        """
        Push the main screen on mount.
        """
        logger.info("Starting application...")

        self.theme = "tokyo-night"

        self.push_screen(screen=screens.ConnectionScreen(config=self.config))

        logger.info("Application startup complete.")



import json
import logging
import pathlib
import sys
from typing import Optional

import dotenv
import typer

import shovl.app as app
import shovl.schemas as schemas
import shovl.services as services

runner = typer.Typer()


logger = logging.getLogger("shovl")


@runner.command()
def run(
    config: Optional[pathlib.Path] = typer.Option(
        None,
        help="Path to configuration file",
    ),
    env: Optional[pathlib.Path] = typer.Option(
        None,
        help="Path to .env file for environment variables",
    ),
    sample_config: bool = typer.Option(
        False,
        help="Print a sample config file",
    ),
    config_schema: bool = typer.Option(
        False,
        help="Print the configuration schema",
    ),
) -> None:
    """
    Run the textual app.
    """
    if env:
        dotenv.load_dotenv(dotenv_path=env)

    if sample_config:
        print(
            services.get_sample_configuration().model_dump_json(indent=4),
            file=sys.stdout,
        )
        return

    if config_schema:
        print(
            json.dumps(schemas.ConfigSchema.model_json_schema(), indent=4),
            file=sys.stdout,
        )
        return

    try:
        config_data = services.get_configuration(path=config)
    except Exception as exception:
        print(f"Error loading configuration: {exception}", file=sys.stderr)
        return

    try:
        services.setup_logging(config=config_data.logging)
    except Exception as exception:
        print(f"Error setting up logging: {exception}", file=sys.stderr)
        return

    try:
        app.ShovlApp(config=config_data).run()
    except Exception as exception:
        print(f"Error running Shovl: {exception}", file=sys.stderr)
        logger.exception(f"Error running Shovl: {exception}")
        return


def main() -> None:
    """
    Serves as an entry point for the application in pyproject.toml.
    """
    runner()


if __name__ == "__main__":
    main()



import logging
from typing import Sequence, Tuple

import sqlalchemy

import shovl.schemas as schemas
import shovl.services as services

from .service import ServiceProvider

logger = logging.getLogger("shovl.providers.database")


class DatabaseProvider(ServiceProvider):
    """
    Database provider for connecting to and inspecting databases using
    SQLAlchemy.
    """

    engine: sqlalchemy.engine.Engine | None
    inspector: sqlalchemy.Inspector | None
    connection: sqlalchemy.Connection | None

    def _cleanup(self) -> None:
        """
        Cleanup the provider's state.
        """
        self.engine = None
        self.inspector = None
        self.connection = None

    @property
    def connected(self) -> bool:
        """
        Check if the provider is connected to a database.
        """
        return (
            self.engine is not None
            or self.inspector is not None
            or self.connection is not None
        )

    def connect(self) -> None:
        """
        Connect to a database.
        """
        logger.info(f"Connecting to database: {self.config.name}")

        assert isinstance(self.config, schemas.DatabaseConfig)

        if self.connected:
            self.disconnect()

        self.engine = sqlalchemy.create_engine(
            url=services.resolve_credentials(self.config.url, parse_url=True)
        )

        self.inspector = sqlalchemy.inspect(self.engine)

        self.connection = self.engine.connect()

        logger.info(f"Successfully connected to database: {self.config.name}")

    def disconnect(self) -> None:
        """
        Disconnect from the database.
        """
        logger.info(f"Disconnecting from database: {self.config.name}")

        if self.connection:
            self.connection.close()

        if self.engine:
            self.engine.dispose()

        self._cleanup()

        logger.info(
            f"Successfully disconnected from database: {self.config.name}"
        )

    def inspect_schemas(self) -> Sequence[schemas.DatabaseEntity]:
        """
        Inspect the database's schemas.
        """
        logger.debug(f"Inspecting schemas in database: {self.config.name}")

        if not self.connected:
            raise Exception("Not connected to a database.")

        assert self.inspector is not None

        names = self.inspector.get_schema_names()

        result = [
            schemas.DatabaseEntity(
                name=name,
                type=schemas.DatabaseEntityType.schema,
            )
            for name in names
        ]

        result = sorted(result, key=lambda x: x.name)

        logger.debug(
            f"Successfully inspected schemas in database: {self.config.name}"
        )

        return result

    def inspect_tables(self, schema: str) -> Sequence[schemas.DatabaseEntity]:
        """
        Inspect a schema's tables.
        """
        logger.debug(f"Inspecting tables in schema '{schema}'.")

        if not self.connected:
            raise Exception("Not connected to a database.")

        assert self.inspector is not None

        names = self.inspector.get_table_names(schema=schema)

        result = [
            schemas.DatabaseEntity(
                name=name,
                type=schemas.DatabaseEntityType.table,
                dbschema=schema,
            )
            for name in names
        ]

        result = sorted(result, key=lambda x: x.name)

        logger.debug(f"Successfully inspected tables in schema '{schema}'.")

        return result

    def inspect_views(self, schema: str) -> Sequence[schemas.DatabaseEntity]:
        """
        Inspect a schema's views.
        """
        logger.debug(f"Inspecting views in schema '{schema}'.")

        if not self.connected:
            raise Exception("Not connected to a database.")

        assert self.inspector is not None

        names = self.inspector.get_view_names(schema=schema)

        result = [
            schemas.DatabaseEntity(
                name=name,
                type=schemas.DatabaseEntityType.view,
                dbschema=schema,
            )
            for name in names
        ]

        result = sorted(result, key=lambda x: x.name)

        logger.debug(f"Successfully inspected views in schema '{schema}'.")

        return result

    def inspect_table(
        self, schema: str, table_name: str
    ) -> Sequence[schemas.DatabaseEntity]:
        """
        Inspect a table, i.e. return its columns.
        """
        logger.debug(f"Inspecting table '{table_name}' in schema '{schema}'.")

        if not self.connected:
            raise Exception("Not connected to a database.")

        assert self.inspector is not None

        columns = self.inspector.get_columns(
            table_name=table_name, schema=schema
        )

        result = []
        for column in columns:
            result.append(
                schemas.DatabaseEntity(
                    name=column["name"],
                    type=schemas.DatabaseEntityType.column,
                    dbschema=schema,
                    datatype=str(column["type"] or ""),
                    nullable=column["nullable"] or True,
                )
            )

        result = sorted(result, key=lambda x: x.name)

        logger.debug(
            f"Successfully inspected table '{table_name}' in schema '{schema}'."
        )

        return result

    def inspect_view(
        self, schema: str, view_name: str
    ) -> Sequence[schemas.DatabaseEntity]:
        """
        Inspect a view, i.e. return its columns.
        """
        logger.debug(f"Inspecting view '{view_name}' in schema '{schema}'.")

        if not self.connected:
            raise Exception("Not connected to a database.")

        assert self.inspector is not None

        columns = self.inspector.get_columns(
            table_name=view_name, schema=schema
        )

        result = []
        for column in columns:
            result.append(
                schemas.DatabaseEntity(
                    name=column["name"],
                    type=schemas.DatabaseEntityType.column,
                    dbschema=schema,
                    datatype=str(column["type"] or ""),
                    nullable=column["nullable"] or True,
                )
            )

        result = sorted(result, key=lambda x: x.name)

        logger.debug(
            f"Successfully inspected view '{view_name}' in schema '{schema}'."
        )

        return result

    def fetch_preview(
        self, schema: str, name: str
    ) -> Sequence[schemas.DatabaseEntity]:
        """
        Fetch a preview of a table's or view's data.
        """
        logger.debug(f"Fetching preview of '{name}' in schema '{schema}'.")

        if not self.connected:
            raise Exception("Not connected to a database.")

        assert self.connection is not None
        assert isinstance(self.config, schemas.DatabaseConfig)

        metadata = sqlalchemy.MetaData()
        table = sqlalchemy.Table(
            name,
            metadata,
            autoload_with=self.engine,
            schema=schema,
        )

        if self.config.max_fetch_rows > 0:
            query = table.select().limit(self.config.max_fetch_rows)
        else:
            query = table.select()

        query_result = self.connection.execute(query)

        rows = [
            {
                k: ("(bytes)" if isinstance(v, bytes) else v)
                for k, v in row.items()
            }
            for row in query_result.mappings()
        ]

        result = [
            schemas.DatabaseEntity(
                name=f"row {index}",
                type=schemas.DatabaseEntityType.row,
                data=row,
            )
            for index, row in enumerate(rows)
        ]

        logger.debug(
            f"Successfully fetched preview of '{name}' in schema '{schema}'."
        )

        return result

    def execute_query(
        self, query: str
    ) -> Tuple[Sequence[schemas.DatabaseEntity], int]:
        """
        Execute a SQL query and return the results as a tuple of a result
        sequence and an affected row count.
        """
        logger.debug(
            f"Executing query: {services.shorten_string(query, max_length=50)}"
        )

        if not self.connected:
            raise Exception("Not connected to a database.")

        assert self.connection is not None

        query_result = self.connection.execute(sqlalchemy.text(query))

        if query_result.returns_rows:
            rows = [
                {
                    k: ("(bytes)" if isinstance(v, bytes) else v)
                    for k, v in row.items()
                }
                for row in query_result.mappings()
            ]

            result = [
                schemas.DatabaseEntity(
                    name=f"row {index}",
                    type=schemas.DatabaseEntityType.row,
                    data=row,
                )
                for index, row in enumerate(rows)
            ]

            affected_rows = 0
        else:
            result = []
            affected_rows = query_result.rowcount or 0

        self.connection.commit()

        logger.debug("Successfully executed query.")

        return (result, affected_rows)



import logging

import shovl.schemas as schemas

from .strings import resolve_path


def setup_logging(config: schemas.LoggingConfig) -> None:
    """
    Set up a file logger for the application.
    """
    logger = logging.getLogger("shovl")

    if not config.enabled:
        logger.addHandler(logging.NullHandler())
        return

    logger.setLevel(config.level)

    file_handler = logging.FileHandler(
        resolve_path(config.file), mode=config.mode, encoding="utf-8"
    )
    file_handler.setLevel(config.level)

    formatter = logging.Formatter("[%(levelname)-10s] %(message)s")

    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)



import copy
import logging
import pathlib

from rich.text import Text
from textual.widgets import DataTable

import shovl.providers as providers
import shovl.schemas as schemas
from shovl.providers.database import DatabaseProvider

from .path import PathScreen
from .service import ServiceScreen

logger = logging.getLogger("shovl.screens.database")


class DatabaseScreen(ServiceScreen):
    """
    Screen to browse the contents of a database.
    """

    BINDINGS = [
        ("backspace", "back", "back"),
        ("f", "filter", "filter"),
        ("p", "preview", "preview"),
        ("y", "query", "query"),
    ]

    last_query_file: pathlib.Path | None

    ### Lifecycle methods ###

    def __init__(
        self,
        config: schemas.ConfigSchema,
        connection: schemas.ServiceConfig,
        provider: providers.ServiceProvider,
    ) -> None:
        """
        Initialize the database screen.
        """
        super().__init__(
            config=config, connection=connection, provider=provider
        )
        self.last_query_file = None

    ### Actions ###

    def action_preview(self) -> None:
        """
        Preview the selected database entity.
        """
        entity = self.get_selected_row()

        assert isinstance(entity, schemas.DatabaseEntity)

        if entity and entity.type in (
            schemas.DatabaseEntityType.table,
            schemas.DatabaseEntityType.view,
        ):
            self.run_in_thread(
                target=self.preview_worker,
                name="PreviewWorker",
                entity=entity,
            )

            self.current_filter = ""
        else:
            self.notify_info("Please select a table or view to preview.")

    def action_query(self) -> None:
        """
        Start a database query using an SQL file.
        """

        def callback(path: pathlib.Path | None) -> None:
            self.last_query_file = path
            if path:
                self.run_in_thread(
                    target=self.query_worker,
                    name="QueryWorker",
                    filepath=path,
                )

        self.app.push_screen(
            PathScreen(
                label="Enter the path to an SQL file:",
                path=self.last_query_file,
                restrictions=[schemas.PathRestriction.must_be_file],
            ),
            callback=callback,
        )

    ### Event handlers ###

    def on_mount(self) -> None:
        """
        Handle the screen mount event by initializing the table, i.e. liting
        availables schemas.
        """
        self.run_in_thread(
            target=self.inspection_worker,
            name="InspectionWorker",
            entity=schemas.DatabaseEntity(
                name="Schemas",
                type=schemas.DatabaseEntityType.schemas,
            ),
            add_current_to_history=False,
        )

    def on_data_table_cell_selected(
        self, event: DataTable.CellSelected
    ) -> None:
        """
        Handle Enter key on list items.
        """
        entity = self.get_selected_row()

        assert isinstance(entity, schemas.DatabaseEntity)

        if entity and entity.type in (
            schemas.DatabaseEntityType.schema,
            schemas.DatabaseEntityType.tables,
            schemas.DatabaseEntityType.views,
            schemas.DatabaseEntityType.table,
            schemas.DatabaseEntityType.view,
        ):
            self.run_in_thread(
                target=self.inspection_worker,
                name="InspectionWorker",
                entity=entity,
                add_current_to_history=True,
            )

    ### Class methods ###

    def inspect(
        self, entity: schemas.ServiceEntity, add_current_to_history: bool
    ) -> None:
        """
        Inspect a specific row in the table. This method is meant to be
        called from a separate thread to avoid blocking the UI.
        """
        assert isinstance(entity, schemas.DatabaseEntity)
        assert isinstance(self.provider, DatabaseProvider)

        match entity.type:
            case schemas.DatabaseEntityType.schemas:
                entities = self.provider.inspect_schemas()
            case schemas.DatabaseEntityType.schema:
                entities = [
                    schemas.DatabaseEntity(
                        name="tables",
                        type=schemas.DatabaseEntityType.tables,
                        dbschema=entity.name,
                    ),
                    schemas.DatabaseEntity(
                        name="views",
                        type=schemas.DatabaseEntityType.views,
                        dbschema=entity.name,
                    ),
                ]
            case schemas.DatabaseEntityType.tables:
                assert entity.dbschema
                entities = self.provider.inspect_tables(schema=entity.dbschema)
            case schemas.DatabaseEntityType.views:
                assert entity.dbschema
                entities = self.provider.inspect_views(schema=entity.dbschema)
            case schemas.DatabaseEntityType.table:
                assert entity.dbschema
                entities = self.provider.inspect_table(
                    table_name=entity.name, schema=entity.dbschema
                )
            case schemas.DatabaseEntityType.view:
                assert entity.dbschema
                entities = self.provider.inspect_view(
                    view_name=entity.name, schema=entity.dbschema
                )
            case _:
                raise Exception(f"Unsupported entity type: {entity.type}")

        self.all_entities = entities
        self.filtered_entities = copy.copy(self.all_entities)

        self.current_filter = ""

        if self.current_entity and add_current_to_history:
            self.add_to_history(self.current_entity)
        self.current_entity = entity

    def fetch_preview(self, entity: schemas.DatabaseEntity) -> None:
        """
        Preview a table or view. Meant to be called from a separate
        thread to avoid blocking the UI.
        """
        assert isinstance(self.provider, DatabaseProvider)
        assert entity.dbschema

        self.all_entities = self.provider.fetch_preview(
            schema=entity.dbschema,
            name=entity.name,
        )
        self.filtered_entities = copy.copy(self.all_entities)

        self.current_filter = ""

        if self.current_entity:
            self.add_to_history(self.current_entity)
        self.current_entity = entity

    def execute_query(self, filepath: pathlib.Path) -> int:
        """
        Execute a query and preview the result. Meant to be called from a
        separate thread to avoid blocking the UI.
        """
        assert isinstance(self.provider, DatabaseProvider)

        if not filepath.exists() or not filepath.is_file():
            raise Exception(f"File not found: {filepath}")

        with filepath.open("r") as file:
            query = file.read()

        entities, affected_rows = self.provider.execute_query(query=query)

        if entities:  # i.e. a select statement
            (self.all_entities, rowcount) = self.provider.execute_query(
                query=query
            )
            self.filtered_entities = copy.copy(self.all_entities)

            self.current_filter = ""

            if self.current_entity:
                self.add_to_history(self.current_entity)

            self.current_entity = None

            return 0
        else:  # i.e. a create/update/insert/... statement
            return affected_rows

    def filter_view(self) -> None:
        """
        Filter database entities based on filter query. Meant to
        be called from a separate thread to avoid blocking the UI.
        """
        query = self.current_filter.lower().strip()

        if not query:
            self.filtered_entities = copy.copy(self.all_entities)
        else:
            self.filtered_entities = []
            for entity in self.all_entities:
                assert isinstance(entity, schemas.DatabaseEntity)
                if not entity.data and query in entity.name.lower():
                    self.filtered_entities.append(entity)
                    continue
                if not entity.data:
                    continue
                if any(
                    query in str(value).lower()
                    for value in entity.data.values()
                ):
                    self.filtered_entities.append(entity)

    def add_columns(self, entity: schemas.ServiceEntity) -> None:
        """
        Add columns to the table based on the type of entity.
        """
        assert isinstance(entity, schemas.DatabaseEntity)

        match entity.type:
            case (
                schemas.DatabaseEntityType.schemas
                | schemas.DatabaseEntityType.schema
                | schemas.DatabaseEntityType.tables
                | schemas.DatabaseEntityType.views
                | schemas.DatabaseEntityType.table
                | schemas.DatabaseEntityType.view
            ):
                self.table.add_columns("Name", "Type")
            case schemas.DatabaseEntityType.column:
                self.table.add_columns("Name", "Data Type", "Nullable", "Type")
            case schemas.DatabaseEntityType.row:
                assert entity.data is not None
                self.table.add_columns(*entity.data.keys())

    def add_row(self, entity: schemas.ServiceEntity) -> None:
        """
        Add a row to the table.
        """
        assert isinstance(entity, schemas.DatabaseEntity)

        name = self.get_entity_display_name(entity=entity)

        primary_color = self.app.theme_variables.get("primary")

        if entity.type in (
            schemas.DatabaseEntityType.schemas,
            schemas.DatabaseEntityType.schema,
            schemas.DatabaseEntityType.tables,
            schemas.DatabaseEntityType.views,
            schemas.DatabaseEntityType.table,
            schemas.DatabaseEntityType.view,
        ):
            self.table.add_row(
                Text(name, style=f"bold {primary_color}"), entity.type.value
            )
        elif entity.type == schemas.DatabaseEntityType.column:
            self.table.add_row(
                name,
                entity.datatype or "",
                "True" if entity.nullable else "False",
                entity.type.value,
            )
        elif entity.type == schemas.DatabaseEntityType.row:
            assert entity.data is not None
            self.table.add_row(*[str(value) for value in entity.data.values()])

    def get_entity_display_name(self, entity: schemas.DatabaseEntity) -> str:
        """
        Get the display name for a database entity.
        """
        return entity.name

    ### Workers ###

    def preview_worker(
        self,
        entity: schemas.DatabaseEntity,
    ) -> None:
        """
        Preview a database table.
        """
        try:
            self.app.call_from_thread(
                self.push_status_screen, label="Fetching preview..."
            )

            self.fetch_preview(entity=entity)

            self.app.call_from_thread(
                self.refresh_view, show_truncated_hint=True
            )

            self.app.call_from_thread(self.dismiss_status_screen)
        except Exception as exception:
            logger.exception(f"Preview failed: {exception}")
            self.app.call_from_thread(self.dismiss_status_screen)
            self.app.call_from_thread(self.focus_view)
            message = str(exception)
            self.app.call_from_thread(
                lambda: self.notify_error(f"Preview failed: {message}")
            )

    def query_worker(
        self,
        filepath: pathlib.Path,
    ) -> None:
        """
        Execute a database query.
        """
        try:
            self.app.call_from_thread(
                self.push_status_screen, label="Executing query..."
            )

            affected_rows = self.execute_query(filepath=filepath)

            if affected_rows:
                self.app.call_from_thread(
                    lambda: self.notify_info(
                        f"Query executed successfully. Affected rows: "
                        f"{affected_rows}"
                    )
                )

            self.app.call_from_thread(
                self.refresh_view, show_truncated_hint=True
            )

            self.app.call_from_thread(self.dismiss_status_screen)
        except Exception as exception:
            logger.exception(f"Query failed: {exception}")
            self.app.call_from_thread(self.dismiss_status_screen)
            self.app.call_from_thread(self.focus_view)
            message = str(exception)
            self.app.call_from_thread(
                lambda: self.notify_error(f"Query failed: {message}")
            )



import copy
import logging
import pathlib

from rich.text import Text
from textual.widgets import DataTable

import shovl.schemas as schemas
import shovl.services as services
from shovl.providers.bucket import BucketProvider

from .path import PathScreen
from .service import ServiceScreen

logger = logging.getLogger("shovl.screens.bucket")


class BucketScreen(ServiceScreen):
    """
    Screen to browse the contents of a S3 bucket.
    """

    BINDINGS = [
        ("backspace", "back", "back"),
        ("f", "filter", "filter"),
        ("d", "download", "download"),
        ("u", "upload", "upload"),
        ("x", "delete", "delete"),
    ]

    ### Actions ###

    def action_download(self) -> None:
        """
        Download the selected bucket entity.
        """
        entity = self.get_selected_row()

        assert isinstance(entity, schemas.BucketEntity)

        if entity and entity.type in (
            schemas.BucketEntityType.file,
            schemas.BucketEntityType.folder,
        ):

            def callback(path: pathlib.Path | None) -> None:
                if path:
                    self.run_in_thread(
                        target=self.download_worker,
                        name="DownloadWorker",
                        entity=entity,
                        target_path=path,
                    )

            self.app.push_screen(
                PathScreen(
                    label="Enter local target directory:",
                    restrictions=[schemas.PathRestriction.must_be_directory],
                ),
                callback=callback,
            )
        else:
            self.notify_info("Only files and folders can be downloaded.")

    def action_upload(self) -> None:
        """
        Upload a local file or directory to the selected bucket or folder.
        """
        entity = self.current_entity

        assert isinstance(entity, schemas.BucketEntity)

        if entity and entity.type in (
            schemas.BucketEntityType.bucket,
            schemas.BucketEntityType.folder,
        ):

            def callback(path: pathlib.Path | None) -> None:
                if path:
                    self.run_in_thread(
                        target=self.upload_worker,
                        name="UploadWorker",
                        entity=entity,
                        source_path=path,
                    )

            self.app.push_screen(
                PathScreen(
                    label="Enter local source file or directory:",
                    restrictions=[schemas.PathRestriction.must_exist],
                ),
                callback=callback,
            )
        else:
            self.notify_info(
                "Please select an S3 bucket or folder as a target."
            )

    def action_delete(self) -> None:
        """
        Delete the selected bucket entity.
        """
        entity = self.get_selected_row()

        assert isinstance(entity, schemas.BucketEntity)

        if entity and entity.type in (
            schemas.BucketEntityType.file,
            schemas.BucketEntityType.folder,
        ):
            self.run_in_thread(
                target=self.delete_worker,
                name="DeleteWorker",
                entity=entity,
            )
        else:
            self.notify_info("Only files and folders can be deleted.")

    ### Event handlers ###

    def on_mount(self) -> None:
        """
        Handle the screen mount event by initializing the table, i.e. listing
        the root bucket contents.
        """
        self.run_in_thread(
            target=self.inspection_worker,
            name="InspectionWorker",
            entity=schemas.BucketEntity(
                name="",
                type=schemas.BucketEntityType.bucket,
            ),
            add_current_to_history=False,
        )

    def on_data_table_cell_selected(
        self, event: DataTable.CellSelected
    ) -> None:
        """
        Handle Enter key on list items.
        """
        entity = self.get_selected_row()

        assert isinstance(entity, schemas.BucketEntity)

        if entity and entity.type == schemas.BucketEntityType.folder:
            self.run_in_thread(
                target=self.inspection_worker,
                name="InspectionWorker",
                entity=entity,
                add_current_to_history=True,
            )

    ### Class methods ###

    def inspect(
        self, entity: schemas.ServiceEntity, add_current_to_history: bool
    ) -> None:
        """
        Inspect a specific row in the table. This method is meant to be
        called from a separate thread to avoid blocking the UI.
        """
        assert isinstance(entity, schemas.BucketEntity)
        assert isinstance(self.provider, BucketProvider)

        match entity.type:
            case schemas.BucketEntityType.bucket:
                entities = self.provider.inspect(prefix=None)
            case schemas.BucketEntityType.folder:
                entities = self.provider.inspect(prefix=entity.name)
            case _:
                raise Exception(f"Unsupported entity type: {entity.type}")

        self.all_entities = entities
        self.filtered_entities = copy.copy(self.all_entities)

        self.current_filter = ""

        if self.current_entity and add_current_to_history:
            self.add_to_history(self.current_entity)
        self.current_entity = entity

    def download(
        self, entity: schemas.BucketEntity, target_path: pathlib.Path
    ) -> None:
        """
        Download a specific bucket entity. Meant to be called from a separate
        thread to avoid blocking the UI.
        """
        assert isinstance(self.provider, BucketProvider)

        if entity.type not in (
            schemas.BucketEntityType.file,
            schemas.BucketEntityType.folder,
        ):
            raise Exception("Only files and folders can be downloaded.")

        self.provider.download(
            key=entity.name,
            target=target_path,
            progress_callback=self.download_progress_callback,
        )

    def upload(
        self, entity: schemas.BucketEntity, source_path: pathlib.Path
    ) -> None:
        """
        Upload local files to bucket. Meant to be called from a separate
        thread to avoid blocking the UI.
        """
        assert isinstance(self.provider, BucketProvider)

        if entity.type not in (
            schemas.BucketEntityType.bucket,
            schemas.BucketEntityType.folder,
        ):
            raise Exception("Only buckets and folders can be upload targets.")

        self.provider.upload(
            key=entity.name,
            source=source_path,
            progress_callback=self.upload_progress_callback,
        )

    def delete(self, entity: schemas.BucketEntity) -> None:
        """
        Upload local files to bucket. Meant to be called from a separate
        thread to avoid blocking the UI.
        """
        assert isinstance(self.provider, BucketProvider)

        if entity.type not in (
            schemas.BucketEntityType.folder,
            schemas.BucketEntityType.file,
        ):
            raise Exception("Only folders and files can be deleted.")

        self.provider.delete(
            key=entity.name,
            progress_callback=self.delete_progress_callback,
        )

    def filter_view(self) -> None:
        """
        Filter bucket entities based on filter query. Meant to
        be called from a separate thread to avoid blocking the UI.
        """
        query = self.current_filter.lower().strip()

        if not query:
            self.filtered_entities = copy.copy(self.all_entities)
        else:
            self.filtered_entities = [
                entity
                for entity in self.all_entities
                if (query in entity.name.lower())
            ]

    def add_columns(self, entity: schemas.ServiceEntity) -> None:
        """
        Add columns to the table based on the type of entity.
        """
        self.table.add_columns("Name", "Size", "Last Modified", "Type")

    def add_row(self, entity: schemas.ServiceEntity) -> None:
        """
        Add a row to the table.
        """
        assert isinstance(entity, schemas.BucketEntity)

        name = self.get_entity_display_name(entity=entity)

        size = services.format_file_size(entity.size) if entity.size else ""
        last_modified = (
            services.to_time_string(entity.last_modified)
            if entity.last_modified
            else ""
        )

        primary_color = self.app.theme_variables.get("primary")

        self.table.add_row(
            Text(
                name,
                style=f"bold {primary_color}"
                if entity.type
                in (
                    schemas.BucketEntityType.folder,
                    schemas.BucketEntityType.bucket,
                )
                else "",
            ),
            size,
            last_modified,
            entity.type.value,
        )

    def get_entity_display_name(self, entity: schemas.BucketEntity) -> str:
        """
        Get the display name for a bucket entity.
        """
        return pathlib.Path(entity.name).name

    ### Callbacks ###

    def download_progress_callback(self, current: int, total: int) -> None:
        """
        Callback function to update the status screen with download progress.
        """
        self.app.call_from_thread(
            lambda: (
                self.status_screen.update_label(
                    f"Downloading file {current}/{total}..."
                )
                if self.status_screen
                else None
            )
        )

    def upload_progress_callback(self, current: int, total: int) -> None:
        """
        Callback function to update the status screen with upload progress.
        """
        self.app.call_from_thread(
            lambda: (
                self.status_screen.update_label(
                    f"Uploading file {current}/{total}..."
                )
                if self.status_screen
                else None
            )
        )

    def delete_progress_callback(self, current: int, total: int) -> None:
        """
        Callback function to update the status screen with delete progress.
        """
        self.app.call_from_thread(
            lambda: (
                self.status_screen.update_label(
                    f"Deleting file {current}/{total}..."
                )
                if self.status_screen
                else None
            )
        )

    ### Workers ###

    def download_worker(
        self, entity: schemas.BucketEntity, target_path: pathlib.Path
    ) -> None:
        """
        Download an entity.
        """
        try:
            self.app.call_from_thread(
                self.push_status_screen, label="Preparing download..."
            )

            self.download(entity=entity, target_path=target_path)

            self.app.call_from_thread(self.dismiss_status_screen)

            self.app.call_from_thread(
                lambda: self.notify_info(
                    f"Download of '{entity.name}' completed."
                )
            )
        except Exception as exception:
            logger.exception(f"Download failed: {exception}")
            self.app.call_from_thread(self.dismiss_status_screen)
            self.app.call_from_thread(self.focus_view)
            message = str(exception)
            self.app.call_from_thread(
                lambda: self.notify_error(f"Download failed: {message}")
            )

    def upload_worker(
        self, entity: schemas.BucketEntity, source_path: pathlib.Path
    ) -> None:
        """
        Upload a path.
        """
        try:
            self.app.call_from_thread(
                self.push_status_screen, label="Preparing upload..."
            )

            self.upload(entity=entity, source_path=source_path)

            if self.current_entity:
                self.inspect(
                    entity=self.current_entity, add_current_to_history=False
                )
                self.app.call_from_thread(
                    self.refresh_view, show_truncated_hint=False
                )

            self.app.call_from_thread(self.dismiss_status_screen)

            self.app.call_from_thread(
                lambda: self.notify_info(
                    f"Upload of '{source_path}' completed."
                )
            )
        except Exception as exception:
            logger.exception(f"Upload failed: {exception}")
            self.app.call_from_thread(self.dismiss_status_screen)
            self.app.call_from_thread(self.focus_view)
            message = str(exception)
            self.app.call_from_thread(
                lambda: self.notify_error(f"Upload failed: {message}")
            )

    def delete_worker(self, entity: schemas.BucketEntity) -> None:
        """
        Delete a path.
        """
        try:
            self.app.call_from_thread(
                self.push_status_screen, label="Preparing deletion..."
            )

            self.delete(entity=entity)

            if self.current_entity:
                self.inspect(
                    entity=self.current_entity, add_current_to_history=False
                )
                self.app.call_from_thread(
                    self.refresh_view, show_truncated_hint=False
                )

            self.app.call_from_thread(self.dismiss_status_screen)

            self.app.call_from_thread(
                lambda: self.notify_info(
                    f"Deletion of '{entity.name}' completed."
                )
            )
        except Exception as exception:
            logger.exception(f"Deletion failed: {exception}")
            self.app.call_from_thread(self.dismiss_status_screen)
            self.app.call_from_thread(self.focus_view)
            message = str(exception)
            self.app.call_from_thread(
                lambda: self.notify_error(f"Deletion failed: {message}")
            )



import datetime
import os
import pathlib
import re
import urllib.parse


def resolve_path(path: str) -> pathlib.Path:
    """
    Resolves a config path by expanding environment variables and user home
    directory.
    """
    path = resolve_credentials(path)
    if path.startswith("~/"):
        resolved_path = pathlib.Path.home() / pathlib.Path(path[2:])
    else:
        resolved_path = pathlib.Path(path)

    return resolved_path.resolve()


def resolve_credentials(template: str, parse_url: bool = False) -> str:
    """
    Resolves placeholders in the given template string with environment
    variable values. Placeholders are in the format {{VAR_NAME}}.
    """
    pattern = re.compile(r"\{\{(\w+)\}\}")

    def replacer(match: re.Match) -> str:
        var_name = match.group(1)
        env_variable = os.getenv(var_name, match.group(0))
        if parse_url:
            env_variable = urllib.parse.quote_plus(env_variable)
        return env_variable

    return pattern.sub(replacer, template)


def shorten_string(text: str, max_length: int = 100) -> str:
    """
    Shortens a string, adding "..." if truncated.
    """
    if len(text) <= max_length:
        return text
    return text[: max_length - 3] + "..."


def to_timestamp(text: str) -> datetime.datetime:
    """
    Converts a string to a datatime object.
    """
    return datetime.datetime.strptime(text, "%Y-%m-%d %H:%M:%S")


def to_time_string(dt: datetime.datetime) -> str:
    """
    Converts a datetime object to a string.
    """
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def format_file_size(bytes: int) -> str:
    """
    Formats a file size in bytes to a human-readable string.
    """
    if bytes < 1024:
        return f"{bytes}B"
    elif bytes < 1024**2:
        return f"{bytes / 1024:.2f}KB"
    elif bytes < 1024**3:
        return f"{bytes / 1024**2:.2f}MB"
    elif bytes < 1024**4:
        return f"{bytes / 1024**3:.2f}GB"
    else:
        return f"{bytes / 1024**4:.2f}TB"



import logging
from typing import List

from textual.app import ComposeResult
from textual.widgets import Footer, Header, Label, ListItem, ListView

import shovl.providers as providers
import shovl.schemas as schemas

from .base import BaseScreen

logger = logging.getLogger("shovl.screens.connection")


class ConnectionScreen(BaseScreen):
    """
    Screen for selecting a connection with filter functionality.
    """

    CSS_PATH = "../styles/connection.tcss"

    BINDINGS = [
        ("f", "filter", "filter"),
        ("q", "quit", "quit"),
    ]

    list_view: ListView

    all_connections: List[schemas.ServiceConfig]
    filtered_connections: List[schemas.ServiceConfig]

    provider: providers.ServiceProvider | None

    ### Lifecycle methods ###

    def __init__(self, config: schemas.ConfigSchema) -> None:
        """
        Initialize the connection screen.
        """
        super().__init__(config=config)

        self.all_connections = self.config.connections
        self.filtered_connections = self.all_connections.copy()

        self.provider = None

        self.set_title("shovl - Select Connection")

    def compose(self) -> ComposeResult:
        """
        Compose the connection selection screen.
        """
        yield Header()
        self.list_view = ListView(
            classes="connection-list", id="connection-list"
        )
        yield self.list_view
        yield Footer()

    ### Actions ###

    def action_quit(self) -> None:
        """
        Quit the app.
        """
        self.app.exit()

    ### Event handlers ###

    def on_mount(self) -> None:
        """
        Initialize the screen when mounted. The connection list is refreshed
        and focused, and if there are any connections, the first one is
        selected.
        """
        self.refresh_view(show_truncated_hint=True)
        self.focus_view()

        if not self.filtered_connections:
            self.notify_info(
                "No connections found. Please add connections to your "
                "config file."
            )

    def on_list_view_selected(self, event: ListView.Selected) -> None:
        """
        Handle Enter key on list items.
        """
        if not self.filtered_connections:
            return

        if (
            self.list_view.index is None
            or self.list_view.index < 0
            or self.list_view.index >= len(self.filtered_connections)
        ):
            return

        selected_connection = self.filtered_connections[self.list_view.index]
        self.run_in_thread(
            target=self.connect_worker,
            name="ConnectWorker",
            connection_config=selected_connection,
        )

    ### Class methods ###

    def filter_view(self) -> None:
        """
        Filter connections based on filter query. Meant to be called from a
        separate thread to avoid blocking the UI.
        """
        query = self.current_filter.lower().strip()

        if not query:
            self.filtered_connections = self.all_connections.copy()
        else:
            self.filtered_connections = [
                conn
                for conn in self.all_connections
                if (
                    query in conn.name.lower()
                    or query in conn.description.lower()
                )
            ]

    def refresh_view(self, show_truncated_hint: bool) -> None:
        """
        Refresh the connection list based on current filter.
        """
        self.list_view.clear()

        for index, connection in enumerate(self.filtered_connections):
            if (
                index >= self.config.gui.max_list_items
                and self.config.gui.max_list_items > 0
            ):
                break

            self.list_view.mount(
                ListItem(
                    Label(
                        connection.name,
                        classes="connection-name",
                    ),
                    Label(
                        connection.description,
                        classes="connection-description",
                    ),
                    classes="connection-item",
                ),
            )

        super().refresh_view(show_truncated_hint=show_truncated_hint)

    def focus_view(self) -> None:
        """
        Focus the connection list and select the first item.
        """
        self.list_view.focus()

        if self.filtered_connections:
            self.list_view.index = 0

    def connect_to_database(
        self,
        connection_config: schemas.DatabaseConfig,
    ) -> None:
        """
        Connect to a database using the provided connection configuration.
        We import the DatabaseProvider here since it depends on optional
        packages. Meant to be called from a separate thread to avoid blocking
        the UI.
        """
        self.test_database_provider()

        from shovl.providers.database import DatabaseProvider

        self.provider = DatabaseProvider(config=connection_config)

        self.provider.connect()

    def test_database_provider(self) -> None:
        """
        Test database provider to ensure they can be imported and used.
        Otherwise, exit gracefully with an error message suggesting how to
        install the required dependencies for database providers.
        """
        try:
            import sqlalchemy

            logger.info(f"SQLAlchemy version is {sqlalchemy.__version__}")
        except Exception:
            raise Exception(
                "Failed to initialize database provider. Try installing "
                "SQLAlchemy."
            )

    def push_database_screen(
        self,
        connection_config: schemas.DatabaseConfig,
    ) -> None:
        """
        Push the database screen upon successful connection. We import the
        DatabaseScreen here since it depends on optional packages.
        """
        from shovl.providers.database import DatabaseProvider
        from shovl.screens.database import DatabaseScreen

        if not isinstance(self.provider, DatabaseProvider):
            raise Exception("Invalid provider type")

        self.dismiss_status_screen()
        self.app.push_screen(
            DatabaseScreen(
                config=self.config,
                connection=connection_config,
                provider=self.provider,
            )
        )

    def connect_to_bucket(
        self,
        connection_config: schemas.BucketConfig,
    ) -> None:
        """
        Connect to an S3 bucket using the provided connection configuration.
        We import the BucketProvider here since it depends on optional
        packages. Meant to be called from a separate thread to avoid blocking
        the UI.
        """
        self.test_bucket_provider()

        from shovl.providers.bucket import BucketProvider

        self.provider = BucketProvider(config=connection_config)

        self.provider.connect()

    def test_bucket_provider(self) -> None:
        """
        Test bucket provider to ensure they can be imported and used.
        Otherwise, exit gracefully with an error message suggesting how to
        install the required dependencies for bucket providers.
        """
        try:
            import boto3
            import botocore

            logger.info(
                f"Boto3 version is {boto3.__version__}, botocore version is "
                f"{botocore.__version__}"
            )
        except Exception:
            raise Exception(
                "Failed to initialize bucket provider. Try installing boto3."
            )

    def push_bucket_screen(
        self,
        connection_config: schemas.BucketConfig,
    ) -> None:
        """
        Push the bucket screen upon successful connection. We import the
        BucketScreen here since it depends on optional packages.
        """
        from shovl.providers.bucket import BucketProvider
        from shovl.screens.bucket import BucketScreen

        if not isinstance(self.provider, BucketProvider):
            raise Exception("Invalid provider type")

        self.dismiss_status_screen()
        self.app.push_screen(
            BucketScreen(
                config=self.config,
                connection=connection_config,
                provider=self.provider,
            )
        )

    ### Workers ###

    def connect_worker(
        self,
        connection_config: schemas.ServiceConfig,
    ) -> None:
        """
        Worker method to connect to the selected service. This is run in a
        seperate thread to avoid blocking the UI.
        """
        try:
            self.app.call_from_thread(
                self.push_status_screen, label="Connecting..."
            )

            if isinstance(connection_config, schemas.DatabaseConfig):
                self.connect_to_database(connection_config=connection_config)
            elif isinstance(connection_config, schemas.BucketConfig):
                self.connect_to_bucket(connection_config=connection_config)

            self.app.call_from_thread(self.dismiss_status_screen)

            if isinstance(connection_config, schemas.DatabaseConfig):
                self.app.call_from_thread(
                    self.push_database_screen, connection_config
                )
            elif isinstance(connection_config, schemas.BucketConfig):
                self.app.call_from_thread(
                    self.push_bucket_screen, connection_config
                )
        except Exception as exception:
            message = str(exception)
            logger.exception(f"Connection failed: {message}")
            self.app.call_from_thread(self.dismiss_status_screen)
            self.app.call_from_thread(self.focus_view)
            self.app.call_from_thread(
                lambda: self.notify_error(f"Unable to connect: {message}")
            )



import enum
from typing import Dict, Optional

import pydantic

from .service import ServiceEntity


class DatabaseEntityType(enum.Enum):
    """
    The type of an entity in the database.
    """

    column = "Column"
    row = "Row"
    schema = "Schema"
    schemas = "Schemas"
    table = "Table"
    tables = "Tables"
    view = "View"
    views = "Views"


class DatabaseEntity(ServiceEntity):
    """
    Represents an entity in a database.
    """

    data: Optional[Dict] = pydantic.Field(
        default=None, description="Data of the row (only for rows)."
    )
    datatype: Optional[str] = pydantic.Field(
        default=None, description="Data type of the column (only for columns)."
    )
    dbschema: Optional[str] = pydantic.Field(
        default=None, description="Parent schema of the entity."
    )
    nullable: Optional[bool] = pydantic.Field(
        default=None,
        description="Whether the column is nullable (only for columns).",
    )
    type: DatabaseEntityType = pydantic.Field(description="Type of the entity.")



import logging
import pathlib
from typing import Callable, Optional, Sequence

import boto3
import botocore.client

import shovl.schemas as schemas
import shovl.services as services

from .service import ServiceProvider

logger = logging.getLogger("shovl.providers.bucket")


class BucketProvider(ServiceProvider):
    """
    Bucket provider for connecting to and inspecting AWS S3 buckets using boto3.
    """

    temp_credentials: schemas.TemporaryCredentials | None

    session: boto3.Session | None
    client: botocore.client.BaseClient | None

    def _cleanup(self) -> None:
        """
        Cleanup the provider's state.
        """
        self.temp_credentials = None
        self.session = None
        self.client = None

    @property
    def connected(self) -> bool:
        """
        Check if the provider is connected to a bucket.
        """
        return self.session is not None or self.client is not None

    def connect(self) -> None:
        """
        Connect to a S3 bucket.
        """
        logger.info(f"Connecting to bucket: {self.config.name}")

        assert isinstance(self.config, schemas.BucketConfig)

        if self.connected:
            self.disconnect()

        self._create_session()

        if self.config.use_sts:
            self._get_temporary_credentials()

        self._create_s3_client()

        assert self.client is not None

        # Optionally test the connection by making a simple API call to the
        # bucket. This can be skipped, e.g. when the user does not have
        # permission to call head_bucket operation.
        if self.config.test_connection:
            self.client.head_bucket(
                Bucket=services.resolve_credentials(self.config.bucket_name)
            )

        logger.info(
            f"Successfully connected to bucket: {self.config.name}. "
            f"Tested connection: {self.config.test_connection}"
        )

    def disconnect(self) -> None:
        """
        Disconnect from a S3 bucket.
        """
        logger.info(f"Disconnecting from bucket: {self.config.name}")

        if self.client:
            self.client.close()

        self._cleanup()

        logger.info(
            f"Successfully disconnected from bucket: {self.config.name}"
        )

    def inspect(
        self, prefix: Optional[str] = None
    ) -> Sequence[schemas.BucketEntity]:
        """
        Inspect the contents of the S3 bucket at the given prefix.
        """
        logger.debug(
            f"Inspecting bucket: {self.config.name} with prefix: {prefix}"
        )

        if not self.connected:
            raise Exception("Not connected to a bucket.")

        assert self.client is not None

        if prefix and not prefix.endswith("/"):
            raise Exception("Prefix must end with a '/' character.")

        assert isinstance(self.config, schemas.BucketConfig)

        # If a prefix is provides, it already contains the head prefix.
        # If not, we need to use it as the prefix. This might be necessary
        # when a user does not have access to the root of the bucket, but
        # only to a specific prefix (e.g. a folder).
        if self.config.head_prefix and (not prefix or len(prefix) == 0):
            prefix = f"{self.config.head_prefix.rstrip('/')}/"

        paginator = self.client.get_paginator("list_objects_v2")

        if prefix:
            page_iterator = paginator.paginate(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Prefix=prefix,
                Delimiter="/",
            )
        else:
            page_iterator = paginator.paginate(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Delimiter="/",
            )

        entities = []
        for page in page_iterator:
            # "Folders" are found in CommonPrefixes
            if "CommonPrefixes" in page:
                for obj in page["CommonPrefixes"]:
                    entities.append(
                        schemas.BucketEntity(
                            name=obj["Prefix"],
                            type=schemas.BucketEntityType.folder,
                        )
                    )

            # Top-level files (if any) are found in 'Contents'
            if "Contents" in page:
                for obj in page["Contents"]:
                    if obj["Key"].endswith("/"):
                        continue
                    entities.append(
                        schemas.BucketEntity(
                            name=obj["Key"],
                            size=obj["Size"],
                            last_modified=obj["LastModified"],
                            type=schemas.BucketEntityType.file,
                        )
                    )

        logger.debug(
            f"Found {len(entities)} entities in bucket: {self.config.name} "
            f"with prefix: {prefix}"
        )
        return entities

    def download(
        self,
        key: str,
        target: pathlib.Path,
        recursive: bool = True,
        progress_callback: Callable[[int, int], None] = lambda x, y: None,
    ) -> None:
        """
        Download a file from the S3 bucket with the given key.
        """
        logger.info(
            f"Downloading from bucket: {self.config.name} with key: {key}"
        )

        if not self.connected:
            raise Exception("Not connected to a bucket.")

        assert self.client is not None
        assert isinstance(self.config, schemas.BucketConfig)

        # List bucket entites
        if recursive:
            entities = self.client.list_objects_v2(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Prefix=key,
            )
        else:
            entities = self.client.list_objects_v2(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Prefix=key,
                Delimiter="/",
            )

        if "Contents" not in entities:
            return

        entities = entities["Contents"]

        # Filter out "folders" (keys that end with "/") and get the list of
        # files to download
        files = [
            entity["Key"]
            for entity in entities
            if "Key" in entity and not entity["Key"].endswith("/")
        ]

        # Get the parent directory of the S3 key. We want to remove the path
        # up to the parent directory from the S3 key when downloading,
        # so that we don't end up with a local directory structure that
        # includes the full S3 key path.
        # "file.ext" -> ""
        # "folder/file.ext" -> "folder"
        # "folder/subfolder/file.ext" -> "folder/subfolder"
        # "folder/" -> ""
        # "folder/subfolder/" -> "folder"
        parent = str(pathlib.Path(key).parent)
        if parent == ".":
            parent = ""

        for index, file in enumerate(files):
            progress_callback(index + 1, len(files))

            relative_path = file[len(parent) :].lstrip("/")
            local_filename = target / relative_path

            if not local_filename.parent.exists():
                local_filename.parent.mkdir(parents=True, exist_ok=True)

            self.client.download_file(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Key=file,
                Filename=str(local_filename),
            )

        logger.info(
            f"Successfully downloaded from bucket: {self.config.name} with "
            f"key: {key} to target: {target}"
        )

    def upload(
        self,
        key: str,
        source: pathlib.Path,
        recursive: bool = True,
        progress_callback: Callable[[int, int], None] = lambda x, y: None,
    ) -> None:
        """
        Upload files to the S3 bucket at the given key.
        """
        logger.info(
            f"Uploading to bucket: {self.config.name} with key: {key} from "
            f"source: {source}"
        )

        if not self.connected:
            raise Exception("Not connected to a bucket.")

        if not source.exists():
            raise Exception(f"Source path '{source}' does not exist.")

        assert self.client is not None
        assert isinstance(self.config, schemas.BucketConfig)

        # Find all local files to upload
        if source.is_file():
            entities = [source]
        elif recursive:
            entities = pathlib.Path(source).rglob("*")
        else:
            entities = pathlib.Path(source).glob("*")

        entities = [entity for entity in list(entities) if entity.is_file()]

        for index, entity in enumerate(list(entities)):
            progress_callback(index + 1, len(entities))

            relative_path = entity.relative_to(source.parent)

            if key:
                s3_key = f"{key.rstrip('/')}/{relative_path.as_posix()}"
            else:
                s3_key = relative_path.as_posix()

            self.client.upload_file(
                Filename=str(entity),
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Key=s3_key,
            )

        logger.info(
            f"Successfully uploaded to bucket: {self.config.name} with key: "
            f"{key} from source: {source}"
        )

    def delete(
        self,
        key: str,
        recursive: bool = True,
        progress_callback: Callable[[int, int], None] = lambda x, y: None,
    ) -> None:
        """
        Delete a file from the S3 bucket with the given key.
        """
        logger.info(f"Deleting from bucket: {self.config.name} with key: {key}")

        if not self.connected:
            raise Exception("Not connected to a bucket.")

        assert self.client is not None
        assert isinstance(self.config, schemas.BucketConfig)

        if recursive:
            entities = self.client.list_objects_v2(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Prefix=key,
            )
        else:
            entities = self.client.list_objects_v2(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Prefix=key,
                Delimiter="/",
            )

        if "Contents" not in entities:
            return

        entities = entities["Contents"]

        files = [
            entity["Key"]
            for entity in entities
            if "Key" in entity and not entity["Key"].endswith("/")
        ]

        for index, file in enumerate(files):
            progress_callback(index + 1, len(files))

            self.client.delete_object(
                Bucket=services.resolve_credentials(self.config.bucket_name),
                Key=file,
            )

        logger.info(
            f"Successfully deleted from bucket: {self.config.name} with key: "
            f"{key}"
        )

    def _create_session(self) -> None:
        """
        Create a boto3 session using the provided configuration.
        """
        logger.info("Creating AWS session...")

        assert isinstance(self.config, schemas.BucketConfig)

        self.session = boto3.Session(
            aws_access_key_id=services.resolve_credentials(
                self.config.aws_access_key_id
            )
            if self.config.aws_access_key_id
            else None,
            aws_secret_access_key=services.resolve_credentials(
                self.config.aws_secret_access_key
            )
            if self.config.aws_secret_access_key
            else None,
            region_name=services.resolve_credentials(self.config.region_name)
            if self.config.region_name
            else None,
            profile_name=services.resolve_credentials(self.config.profile_name)
            if self.config.profile_name
            else None,
            aws_account_id=services.resolve_credentials(
                self.config.aws_account_id
            )
            if self.config.aws_account_id
            else None,
        )

        logger.info("AWS session created successfully.")

    def _create_s3_client(self) -> None:
        """
        Create a boto3 bucket resource using the session.
        """
        logger.info("Creating S3 client...")

        if not self.session:
            raise Exception("AWS session not created.")

        assert isinstance(self.config, schemas.BucketConfig)

        if not self.config.use_sts:
            self.client = self.session.client("s3")
        else:
            if not self.temp_credentials:
                raise Exception("Temporary credentials not obtained.")

            self.client = boto3.client(
                "s3",
                aws_access_key_id=self.temp_credentials.access_key_id,
                aws_secret_access_key=self.temp_credentials.secret_access_key,
                aws_session_token=self.temp_credentials.session_token,
                region_name=services.resolve_credentials(
                    self.config.region_name
                )
                if self.config.region_name
                else None,
            )

        logger.info("S3 client created successfully.")

    def _get_temporary_credentials(self) -> None:
        """
        Get temporary credentials for accessing the S3 bucket.
        """
        logger.debug("Fetching temporary credentials using AWS STS...")

        if not self.session:
            raise Exception("AWS session not created.")

        assert isinstance(self.config, schemas.BucketConfig)

        account_id = (
            services.resolve_credentials(self.config.aws_account_id)
            if self.config.aws_account_id
            else None
        )

        role_name = (
            services.resolve_credentials(self.config.role_name)
            if self.config.role_name
            else None
        )

        sts = self.session.client("sts")
        response = sts.assume_role(
            RoleArn=f"arn:aws:iam::{account_id}:role/{role_name}",
            RoleSessionName="ShovlBucketProviderSession",
        )

        if "Credentials" not in response:
            raise Exception(
                "Failed to assume AWS role for temporary credentials."
            )

        self.temp_credentials = schemas.TemporaryCredentials(
            **response["Credentials"]
        )

        logger.debug("Temporary credentials obtained successfully.")



import logging
import pathlib
import threading
from typing import Callable, List, Optional

from textual.app import ComposeResult
from textual.containers import Vertical
from textual.events import Key
from textual.screen import ModalScreen
from textual.widgets import Input, Label

import shovl.schemas as schemas

logger = logging.getLogger("shovl.screens")


class PathScreen(ModalScreen[pathlib.Path | None]):
    """
    Screen to enter a local file path.
    """

    CSS_PATH = "../styles/path.tcss"

    label: Label
    hint: Label

    worker: threading.Thread | None
    path_changed_event: threading.Event

    initial_label: str | None
    initial_path: pathlib.Path | None

    current_path: str
    restrictions: List[schemas.PathRestriction]

    ### Lifecycle methods ###

    def __init__(
        self,
        label: Optional[str] = None,
        path: Optional[pathlib.Path] = None,
        restrictions: List[schemas.PathRestriction] = [],
    ) -> None:
        """
        Initialize the status screen.
        """
        super().__init__()

        self.initial_label = label
        self.initial_path = path

        self.worker = None
        self.path_changed_event = threading.Event()

        self.current_path = str(path) if path else ""
        self.restrictions = restrictions

    def compose(self) -> ComposeResult:
        """
        Compose the status screen.
        """
        with Vertical(classes="path-container"):
            label = (
                self.initial_label if self.initial_label else "Enter local path"
            )
            path = (
                self.initial_path if self.initial_path else pathlib.Path.cwd()
            )
            self.current_path = str(path)
            self.label = Label(label, classes="label")
            yield self.label
            yield Input(classes="input", value=self.current_path)
            self.hint = Label("Checking path...", classes="hint")
            yield self.hint

    ### Event handlers ###

    def on_mount(self) -> None:
        """
        Start the path check worker when the screen is mounted.
        """
        self.run_in_thread(
            self.path_check_worker,
            name="PathCheckWorker",
            callback=self.path_checked_callback,
        )

    def on_input_changed(self, event: Input.Changed) -> None:
        """
        Handle changes to the input field.
        """
        self.current_path = event.value
        self.path_changed_event.set()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        """
        Handle submission of the input field. Closes the screen if the path is
        valid.
        """
        if self.check_path(self.current_path):
            self.dismiss(pathlib.Path(self.current_path))

    def on_key(self, event: Key) -> None:
        """
        Handle key presses within the input field. Closes the screen if the
        user hits the escape key.
        """
        if event.key == "escape":
            self.dismiss(None)

    ### Callbacks ###

    def path_checked_callback(self, valid: bool) -> None:
        """
        Updates the label indicating whether the path is valid or not.
        """
        if valid:
            self.hint.update(content="✅ Path is valid")
        else:
            self.hint.update(content="🛑 Path is not valid")

    ### Class methods ###

    def run_in_thread(
        self, target: Callable, name: str, *args, **kwargs
    ) -> None:
        """
        Run the given target function in a separate thread. Shows the status
        screen while the thread is running.
        """
        if self.worker and self.worker.is_alive():
            raise Exception("Worker is already running")

        self.worker = threading.Thread(
            target=target, daemon=True, name=name, args=args, kwargs=kwargs
        )
        self.worker.start()

    def check_path(self, path: pathlib.Path | str) -> bool:
        """
        Return True is a path is an existing directory or False otherwise.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)

        if (
            schemas.PathRestriction.must_exist in self.restrictions
            and not path.exists()
        ):
            return False

        if (
            schemas.PathRestriction.must_be_directory in self.restrictions
            and not path.is_dir()
        ):
            return False

        if (
            schemas.PathRestriction.must_be_file in self.restrictions
            and not path.is_file()
        ):
            return False

        return True

    ### Worker methods ###

    def path_check_worker(
        self,
        callback: Callable[[bool], None] = lambda x: None,
    ) -> None:
        """
        Check if the path is valid.
        """
        while True:
            self.path_changed_event.wait()

            self.app.call_from_thread(
                callback, self.check_path(path=self.current_path)
            )

            self.path_changed_event.clear()



from typing import Optional

from textual.app import ComposeResult
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Label, LoadingIndicator


class StatusScreen(ModalScreen):
    """
    Screen to display a status.
    """

    CSS_PATH = "../styles/status.tcss"

    label: Label

    initial_label: str | None

    ### Lifecycle methods ###

    def __init__(self, label: Optional[str] = None) -> None:
        """
        Initialize the status screen.
        """
        super().__init__()

        self.initial_label = label

    def compose(self) -> ComposeResult:
        """
        Compose the status screen.
        """
        with Vertical(classes="status-container"):
            label = self.initial_label if self.initial_label else "Loading..."
            self.label = Label(label, classes="label")
            yield self.label
            yield LoadingIndicator(classes="indicator")

    ### Class methods ###

    def update_label(self, label: str) -> None:
        """
        Update label.
        """
        self.label.update(content=label)



from .config import (
    get_configuration,
    get_sample_configuration,
)
from .logging import setup_logging
from .strings import (
    format_file_size,
    resolve_credentials,
    resolve_path,
    shorten_string,
    to_time_string,
    to_timestamp,
)

__all__ = [
    "get_configuration",
    "get_sample_configuration",
    "setup_logging",
    "format_file_size",
    "resolve_credentials",
    "resolve_path",
    "shorten_string",
    "to_time_string",
    "to_timestamp",
]



import pydantic


class ServiceEntity(pydantic.BaseModel):
    """
    Base entity model for services. Represents a generic entity in a service,
    such as a file, folder, or table. This model can be extended by specific
    service models to include additional fields relevant to that service.
    """

    name: str = pydantic.Field(description="Name of the entity.")



from typing import List, Optional, Union

import pydantic


class GuiConfig(pydantic.BaseModel):
    max_list_items: int = pydantic.Field(
        default=1000,
        description="Maximum number of entities to display in the UI. Set to "
        "0 for no limit.",
    )


class LoggingConfig(pydantic.BaseModel):
    enabled: bool = pydantic.Field(
        default=False, description="Enable or disable logging."
    )
    file: str = pydantic.Field(
        default="~/.shovl.log", description="Path to the log file."
    )
    level: str = pydantic.Field(
        default="INFO",
        description="Logging level. One of: DEBUG, INFO, WARNING, ERROR, "
        "CRITICAL.",
    )
    mode: str = pydantic.Field(
        default="w",
        description="File mode for logging. 'a' for append, 'w' for overwrite.",
    )


class ProviderConfig(pydantic.BaseModel):
    description: str = pydantic.Field(
        description="Short description of the connection."
    )
    name: str = pydantic.Field(description="Name of the connection.")


class DatabaseConfig(ProviderConfig):
    max_fetch_rows: int = pydantic.Field(
        default=1000,
        description="Maximum number of rows to fetch from the database. Set to "
        "0 for no limit.",
    )
    url: str = pydantic.Field(description="Database connection URL.")


class BucketConfig(ProviderConfig):
    aws_access_key_id: Optional[str] = pydantic.Field(
        default=None, description="AWS access key ID for S3 access."
    )
    aws_account_id: Optional[str] = pydantic.Field(
        default=None, description="AWS account ID for STS access."
    )
    aws_secret_access_key: Optional[str] = pydantic.Field(
        default=None, description="AWS secret access key for S3 access."
    )
    bucket_name: str = pydantic.Field(description="Name of the S3 bucket.")
    head_prefix: Optional[str] = pydantic.Field(
        default=None,
        description="General prefix to use when listing objects in the bucket.",
    )
    profile_name: Optional[str] = pydantic.Field(
        default=None, description="AWS profile name for S3 access."
    )
    region_name: Optional[str] = pydantic.Field(
        default=None, description="AWS region name for S3 access."
    )
    role_name: Optional[str] = pydantic.Field(
        default=None, description="AWS role name for STS access."
    )
    test_connection: bool = pydantic.Field(
        default=True,
        description="Whether to test the connection to the S3 bucket on "
        "startup.",
    )
    use_sts: bool = pydantic.Field(
        default=False,
        description="Whether to use AWS Security Token Service for temporary "
        "credentials. Requires a role name and AWS account ID if set to True.",
    )


ServiceConfig = Union[DatabaseConfig, BucketConfig]


class ConfigSchema(pydantic.BaseModel):
    connections: List[ServiceConfig] = pydantic.Field(
        default=[],
        description="List of connections to data sources.",
    )
    gui: GuiConfig = pydantic.Field(
        default=GuiConfig(), description="GUI configuration."
    )
    logging: LoggingConfig = pydantic.Field(
        default=LoggingConfig(), description="Logging configuration."
    )



import logging
import threading
from typing import Callable, Optional

from textual.screen import Screen

import shovl.schemas as schemas
import shovl.services as services

from .filter import FilterScreen
from .status import StatusScreen

logger = logging.getLogger("shovl.screens.base")


class BaseScreen(Screen):
    """
    Base screen. Inherited by all other screens. Offers common functionality
    like status and filter handling.
    """

    config: schemas.ConfigSchema

    status_screen: StatusScreen | None
    filter_screen: FilterScreen | None

    worker: threading.Thread | None

    current_filter: str
    filter_changed_signal: threading.Event
    filter_closed_signal: threading.Event

    refreshed_signal: threading.Event

    ### Lifecycle methods ###

    def __init__(self, config: schemas.ConfigSchema) -> None:
        """
        Initialize the service screen.
        """
        super().__init__()

        self.config = config

        self.status_screen = None
        self.filter_screen = None

        self.worker = None

        self.current_filter = ""
        self.filter_changed_signal = threading.Event()
        self.filter_closed_signal = threading.Event()

        self.refreshed_signal = threading.Event()

    ### Actions ###

    def action_filter(self) -> None:
        """
        Start the filter worker showing the filter screen.
        """
        self.run_in_thread(self.filter_worker, name="FilterWorker")

    ### Callbacks ###

    def filter_change_callback(self, query: str) -> None:
        """
        Callback for the filter modal. Called anytime the input value changes.
        """
        self.current_filter = query
        self.filter_changed_signal.set()

    def filter_closed_callback(self, query: str) -> None:
        """
        Callback for the filter modal. Called when the user hits the enter key
        within the filter modal.
        """
        self.current_filter = query
        self.filter_closed_signal.set()
        self.filter_changed_signal.set()

    ### Class methods ###

    def set_title(self, title: str) -> None:
        """
        Set the title of the screen.
        """
        self.title = services.shorten_string(title, 30)

    def notify_error(self, message: str) -> None:
        """
        Notify the user of an error.
        """
        logger.debug(f"Notifying user of an error: {message}")
        self.app.notify(
            services.shorten_string(message, 200),
            title="Error",
            severity="error",
        )

    def notify_info(self, message: str) -> None:
        """
        Notify the user of an informational message.
        """
        logger.debug(f"Notifying user of an informational message: {message}")
        self.app.notify(
            services.shorten_string(message, 200),
            title="Info",
            severity="information",
        )

    def run_in_thread(
        self, target: Callable, name: str, *args, **kwargs
    ) -> None:
        """
        Run the given target function in a separate thread. Shows the status
        screen while the thread is running.
        """
        if self.worker and self.worker.is_alive():
            raise Exception("Worker is already running.")

        self.worker = threading.Thread(
            target=target, daemon=True, name=name, args=args, kwargs=kwargs
        )
        self.worker.start()

    def push_status_screen(self, label: Optional[str] = None) -> None:
        """
        Push the status screen.
        """
        if self.status_screen:
            raise Exception("Status screen is already being shown.")

        self.status_screen = StatusScreen(label=label)
        self.app.push_screen(self.status_screen)

    def dismiss_status_screen(self) -> None:
        """
        Dismiss the status screen.
        """
        if self.status_screen:
            self.status_screen.dismiss()
            self.status_screen = None

    def push_filter_screen(self) -> None:
        """
        Show the modal filter screen.
        """
        if self.filter_screen:
            raise Exception("Filter screen is already being shown.")

        self.filter_changed_signal.clear()
        self.filter_closed_signal.clear()
        self.refreshed_signal.clear()

        self.filter_screen = FilterScreen(
            current_filter=self.current_filter,
            change_callback=self.filter_change_callback,
            confirm_callback=self.filter_closed_callback,
            cancel_callback=self.filter_closed_callback,
        )
        self.app.push_screen(self.filter_screen)

    def dismiss_filter_screen(self) -> None:
        """
        Dismiss the modal filter screen.
        """
        if self.filter_screen:
            self.filter_screen.dismiss()
            self.filter_screen = None

    def filter_view(self) -> None:
        """
        Handle the filter operation for the default filter worker. Has to be
        implemented by the subclass.
        """
        raise NotImplementedError(
            "Filter method must be implemented by subclass."
        )

    def refresh_view(self, show_truncated_hint: bool) -> None:
        """
        Handle the refresh operation for the default filter worker. Has to be
        implemented by the subclass. Must call super method!
        """
        self.call_after_refresh(self.refreshed_signal.set)

    def focus_view(self) -> None:
        """
        Focus the main view after the default filter worker has been dismissed.
        Has to be implemented by the subclass.
        """
        raise NotImplementedError(
            "Focus main method must be implemented by subclass."
        )

    ### Workers ###

    def filter_worker(self) -> None:
        """
        Default implementation for a filter worker. Shows the filter screen and
        waits for the filter signal to be set. When the signal is set, it
        checks the filter status and either cancels, confirms, or updates the
        filter results. This loop continues until the filter is either confirmed
        or cancelled.
        """
        try:
            self.app.call_from_thread(self.push_filter_screen)

            while True:
                self.filter_changed_signal.wait()

                self.filter_view()
                self.app.call_from_thread(
                    self.refresh_view, show_truncated_hint=False
                )

                # We wait after the refresh is done before we potentially
                # update the filter again, because the refresh might take some
                # time and we don't want to update the filter results while the
                # refresh is still running to avoid potential race conditions.
                self.refreshed_signal.wait()
                self.refreshed_signal.clear()

                if self.filter_closed_signal.is_set():
                    break

                self.filter_changed_signal.clear()
        except Exception as exception:
            logger.exception(f"Error in filter worker: {exception}")
            message = str(exception)
            self.app.call_from_thread(
                lambda: self.notify_error(
                    f"Unable to filter connections: {message}"
                )
            )

        self.app.call_from_thread(self.dismiss_filter_screen)
        self.app.call_from_thread(self.focus_view)



import functools
import json
import pathlib
from typing import Optional

import shovl.schemas as schemas


@functools.lru_cache()
def get_configuration(
    path: Optional[pathlib.Path] = None,
) -> schemas.ConfigSchema:
    """
    Read the configuration from disk and return it as a ConfigSchema object.
    """
    default_paths = [
        pathlib.Path(".shovl"),
        pathlib.Path.home() / pathlib.Path(".shovl"),
    ]

    if not path:
        for default_path in default_paths:
            if default_path.is_file():
                path = default_path
                break

    if not path:
        return schemas.ConfigSchema()

    with path.open("r") as file:
        config_data = json.load(file)

    return schemas.ConfigSchema(**config_data)


def get_sample_configuration() -> schemas.ConfigSchema:
    """
    Create a sample configuration.
    """
    config = schemas.ConfigSchema()

    config.connections.append(
        schemas.DatabaseConfig(
            name="Sample SQLite Database",
            description="A sample SQLite connection",
            url="sqlite:///sample.db",
        )
    )

    config.connections.append(
        schemas.BucketConfig(
            name="Sample S3 Bucket",
            description="A sample S3 bucket connection",
            bucket_name="shovl-app-test-bucket",
            aws_access_key_id="{{TEST_BUCKET_ACCESS_KEY_ID}}",
            aws_secret_access_key="{{TEST_BUCKET_SECRET_ACCESS_KEY}}",
            region_name="{{TEST_BUCKET_REGION_NAME}}",
            aws_account_id="{{TEST_BUCKET_ACCOUNT_ID}}",
        )
    )

    return config
